/**********************************************************************
 * Copyright (c) 2007 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.content;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import org.eclipse.core.runtime.QualifiedName;
import org.eclipse.core.runtime.content.IContentDescriber;
import org.eclipse.core.runtime.content.IContentDescription;

/**
 * The abstract XML Content describer provides generic mechanisms for
 * determining the content of an XML document
 * 
 * @author Ali Mehregani
 */
public abstract class AbstractXMLContentDescriber implements IContentDescriber
{
	/**
	 * Indicates that the content search should continue 
	 */
	private static final int CONTINUE = 10;
	
	/**
	 * @see org.eclipse.core.runtime.content.IContentDescriber#describe(java.io.InputStream, org.eclipse.core.runtime.content.IContentDescription)
	 */
	public int describe(InputStream contents, IContentDescription description) throws IOException
	{
		BufferedReader br = new BufferedReader(new InputStreamReader(contents));
		
		String currentLine = null;
		boolean commentHit = false;
		int inx = -1;
		while ((currentLine = br.readLine()) != null)
		{
			currentLine = currentLine.trim();
			if (currentLine.length() <= 0 || currentLine.startsWith("<?"))
			{
				continue;
			}
			else if (currentLine.startsWith("<!--"))
			{
				if (currentLine.endsWith("-->"))
					continue;
				commentHit = true;
			}
			else if (commentHit && (inx = currentLine.indexOf("-->")) >= 0)
			{
				commentHit = false;
				if (inx < currentLine.length() - 4)
				{
					currentLine = currentLine.substring(inx).trim();
					int type = determineType(currentLine);
					if (type == CONTINUE)
						continue;
					return type;
				}
			}
			else if (!commentHit)				
			{
				int type = determineType(currentLine);
				if (type == CONTINUE)
					continue;
				return type;

			}
		}
		
		return IContentDescriber.INVALID;
	}

	
	/**
	 * Used to determine the type of the content based on the line passed in
	 * 
	 * @param line The line being processed
	 */
	public abstract int determineType(String line);


	/**
	 * @see org.eclipse.core.runtime.content.IContentDescriber#getSupportedOptions()
	 */
	public QualifiedName[] getSupportedOptions()
	{		
		return null;
	}
	
	
	protected String retrieveElementName (String line)
	{
		if (line == null || line.length() <= 0 || line.charAt(0) != '<')
			return null;
		
		int inx = line.indexOf(' ');
		line = inx < 0 ? line.substring(1) : line.substring(1, inx);
		inx = line.indexOf(':');
		return inx < 0 ? line : line.substring(inx + 1);
	}

}
