/*******************************************************************************
 * Copyright (c) 2003, 2006 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.cosmos.rm.validation.internal.databuilders;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.eclipse.cosmos.rm.validation.internal.SMLActivator;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.eclipse.cosmos.rm.validation.internal.common.SMLMessages;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.xml.sax.Attributes;

/**
 * This class will build a map containing all phenic documents in a DOM format.
 * The documents will be identified by the document alias
 * 
 * @author Ali Mehregani
 */
public class DocumentDOMBuilder extends AbstractDataBuilder
{
	/**
	 * The ID of this builder
	 */
	public static final String ID = SMLActivator.PLUGIN_ID + ".XMLDefaultHandler";
	
	/**
	 * The key used for documents that don't have aliases
	 */
	public static final String NO_ALIAS = "no_alias";
	
	/**
	 * The root document
	 */
	private Document document;
	
	/**
	 * Stores the document nodes
	 * 
	 * KEY = A string representing a document alias
	 * VALUE = An object of type {@link Node} representing the root document node
	 */
	private Map domDocumentMap;

	/**
	 * Stores a list of documents nodes without an alias
	 */
	private List documentsWithNoAlias;
	
	/**
	 * The current root element of the document
	 */
	private Node rootElement;
	
	/**
	 * Keeps track of the elements of the current document element
	 */
	private Stack elementStack = new Stack();
	
	/**
	 * The current alias list
	 */
	private List aliasesList;
	
	/**
	 * Current alias
	 */
	private String currentAlias;
	
	/**
	 * Indicates whether a CDATA section is reached 
	 */
	private boolean cdataStarted;

	/**
	 * Indicate that the alias element is hit
	 */
	private boolean aliasElementHit;
	
	/**
	 * Indicates that the document element is hit
	 */
	private boolean documentElementHit;
	
	/**
	 * Indicates that the data element is hit
	 */
	private boolean dataElementHit;

	/**
	 * Indicates that the phenic element is hit
	 */
	private boolean phenicElementHit;
	
	public DocumentDOMBuilder()
	{
		domDocumentMap = new Hashtable();
		documentsWithNoAlias = new ArrayList();
		domDocumentMap.put(NO_ALIAS, documentsWithNoAlias);
		currentAlias = IValidationConstants.EMPTY_STRING;
		aliasesList = new ArrayList();
	}

	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.AbstractDataBuilder#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uri, String localName, String qName, Attributes attributes)
	{		
		/* The document and data element is hit */
		if (dataElementHit)
		{			
			if (document == null)
			{
				DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				factory.setNamespaceAware(true);
				factory.setIgnoringElementContentWhitespace(true);
				
				try
				{			
					document = factory.newDocumentBuilder().newDocument();			
				}
				catch (ParserConfigurationException e)
				{
					setErrorMessage(SMLMessages.errorCreatingDocument);
					setStructureValidity(false);
					return;
				}
			}
			
			Element element = document.createElementNS(uri, qName);
			for (int i = 0; i < attributes.getLength(); i++)
			{
				element.setAttribute(attributes.getQName(i), attributes.getValue(i));
			}
			
			if (rootElement == null)
			{
				rootElement = element;
			}
			else
			{
				((Element) elementStack.peek()).appendChild(element);
			}
			elementStack.push(element);
		}
		
		
		if (ISMLConstants.SMLIF_URI.equals(uri))
		{
			/* The alias element is hit */
			if (ISMLConstants.ALIAS_ELEMENT.equals(localName))
			{
				aliasElementHit = true;
			}
			/* This test is only needed because of the associated JUnit */
			else if (ISMLConstants.PHENICS_ELEMENT.equals(localName))
			{
				phenicElementHit = true;
			}
			/* The document element is hit */
			else if (phenicElementHit && ISMLConstants.DOCUMENT_ELEMENT.equals(localName))
			{
				documentElementHit = true;				
			}
			/* The data element is hit */
			else if (documentElementHit && ISMLConstants.DATA_ELEMENT.equals(localName))
			{
				dataElementHit = true;
			}
		}
	}

	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.AbstractDataBuilder#endElement(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void endElement(String uri, String localName, String qName)
	{
		if (ISMLConstants.SMLIF_URI.equals(uri))
		{
			/* The alias element is hit */
			if (ISMLConstants.ALIAS_ELEMENT.equals(localName))
			{
				if (currentAlias.length() >  0)
					aliasesList.add(currentAlias);
				currentAlias = IValidationConstants.EMPTY_STRING;
				aliasElementHit = false;
			}
			else if (phenicElementHit && ISMLConstants.PHENICS_ELEMENT.equals(localName))
			{
				phenicElementHit = false;				
			}
			/* The document element is hit */
			else if (documentElementHit && ISMLConstants.DOCUMENT_ELEMENT.equals(localName))
			{
				documentElementHit = false;				
			}
			/* The data element is hit */
			else if (dataElementHit && ISMLConstants.DATA_ELEMENT.equals(localName))
			{
				dataElementHit = false;
				
				/* Iterate through the aliases and add them to the map */
				if (rootElement != null)
				{
					document.appendChild(rootElement);
					if (aliasesList.size() <= 0)
						documentsWithNoAlias.add(rootElement);
					
					for (int i = 0, aliasCount = aliasesList.size(); i < aliasCount; i++)
					{
						domDocumentMap.put(aliasesList.get(i), rootElement);
					}
				}
				aliasesList.clear();
				rootElement = null;
				document = null;
			}
		}
		
		
		if (dataElementHit)
		{
			elementStack.pop();		
		}
	}


	
	/**
	 * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
	 */
	public void characters(char[] characters, int start, int length)
	{		
		if (aliasElementHit)
		{
			currentAlias += new String(characters, start, length);
		}
		else if (dataElementHit)
		{		
			if (rootElement == null)
				return;
			
			String buffValue = new String(characters, start, length);
			boolean isBufferJustWhiteSpace = IValidationConstants.EMPTY_STRING.equals(buffValue.trim());
			boolean isBufferEmpty = buffValue.length() <= 0;
			
			if ((cdataStarted && isBufferEmpty) || (!cdataStarted && isBufferJustWhiteSpace))
			{
				cdataStarted = false;
				return;
			}
					
			Text text = document.createTextNode(buffValue);
			cdataStarted = true;
			
			if (elementStack.size() <= 0)
			{
				rootElement.appendChild(text);
			}
			else
			{
				((Node) elementStack.peek()).appendChild(text);
			}
		}
	}
	
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#getDataStructure()
	 */
	public Object getDataStructure()
	{
		return domDocumentMap;
	}
	
	public List getDocumentsWithNoAlias(){
		return documentsWithNoAlias;
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.AbstractDataBuilder#getPhase()
	 */
	public byte getPhase()
	{
		return ISMLConstants.UNKNOWN_PHASE;
	}
}
