/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.reference;

import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.xml.xpath.XPathFunction;
import javax.xml.xpath.XPathFunctionException;

import org.eclipse.cosmos.rm.validation.internal.artifacts.TypeNode;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil;
import org.eclipse.cosmos.rm.validation.internal.common.SMLMessages;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil.RemoteRetrievalException;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DocumentDOMBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IdentityDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RefTypeElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TypeInheritanceDataBuilderImpl;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IdentityDataBuilder.SMLIFIdentity;
import org.eclipse.osgi.util.NLS;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This is an implementaion of the dref function specified in the SML
 * specification.  This implementation makes the following assumptions:
 * <ul>
 * 	<li> The function is invoked after parsing the documents to be validated (i.e.
 * 	the data builders have constructed their associated data structure) 
 * 	</li>
 * 	<li> The data structure of the data builder with the ID 
 * 	{@link DocumentDOMBuilder#ID} exists.
 * 	</li>
 * </ul>
 * This is a singleton class that should be accessed via {@link DerefXPathFunction#instance()} 
 *  
 * @author Ali Mehregani
 */
public class DerefXPathFunction implements XPathFunction
{
	/**
	 * Arity - The number of arguments for the dref function
	 */
	public static final int ARITY = 1;
	
	/**
	 * The instance of this singleton class
	 */
	private static DerefXPathFunction instance;
	
	
	/**
	 * @see javax.xml.xpath.XPathFunction#evaluate(java.util.List)
	 */
	public Object evaluate(List arguments) throws XPathFunctionException
	{
		/* Retrieve the data structures required */
		Map refTypeElements = (Map)SMLCommonUtil.retrieveDataStructure(RefTypeElementBuilder.ID);
		Map domDocuments = (Map) SMLCommonUtil.retrieveDataStructure(DocumentDOMBuilder.ID);		
		SMLIFIdentity identity = (SMLIFIdentity) SMLCommonUtil.retrieveDataStructure(IdentityDataBuilder.ID);
		Map inheritanceMap = (Map) SMLCommonUtil.retrieveDataStructure(TypeInheritanceDataBuilderImpl.ID);
		inheritanceMap = inheritanceMap == null ? new Hashtable() : inheritanceMap;
		
		if (refTypeElements == null || domDocuments == null || identity == null)
			throw new XPathFunctionException(SMLMessages.errorMissingStructure);
		
		NodeList nodeList = (NodeList)arguments.get(0);
		ArrayNodeList result = new ArrayNodeList();
		Exception exception = null;
		
		/* For every node in the node list */
		for (int i = 0, nodeCount = nodeList.getLength(); i < nodeCount; i++)
		{
			Node currentNode = nodeList.item(i);
			
			/* Determine the uri and name of the node */
			String name = currentNode.getNodeName();
			String uri = currentNode.getNamespaceURI();
			uri = uri == null ? IValidationConstants.EMPTY_STRING : uri;						
			if (name == null)
				throw new XPathFunctionException(SMLMessages.errorDetermineName);
			
			/* Determine if the node is of type 'refType' or derived from a type that is of type 'refType' */
			TypeNode currentTypeNode = new TypeNode (name, uri);
			Map typeMap = null;
			boolean refType = false;
			while (currentTypeNode != null && !refType)
			{
				refType = (typeMap= (Map)refTypeElements.get(currentTypeNode.getUri())) != null && typeMap.get(currentTypeNode.getType()) != null;
				currentTypeNode = TypeInheritanceDataBuilderImpl.retrieveType(inheritanceMap, currentTypeNode.getUri(), currentTypeNode.getType());
			}
			
			if (!refType)
				throw new XPathFunctionException(NLS.bind(SMLMessages.errorWrongType, name));
			

			
			/* Find the reference URI */
			NodeList currentChildNodes = currentNode.getChildNodes();
			if (nodeList == null)
			{
				continue;
			}
			
			StringBuffer referenceExpression = new StringBuffer();
			for (int j = 0, childNodeCount = currentChildNodes.getLength(); j < childNodeCount; j++)
			{
				Node currentChildNode = currentChildNodes.item(j);
				if (ISMLConstants.SML_URI.equals(currentChildNode.getNamespaceURI()) && ISMLConstants.URI_ELEMENT.equals(currentChildNode.getLocalName()))
				{
					NodeList uriNodes = currentChildNode.getChildNodes();
					for (int k = 0, uriNodeCount = uriNodes.getLength(); k < uriNodeCount; k++)
					{
						Node uriNode = uriNodes.item(k);
						if (Node.TEXT_NODE == uriNode.getNodeType() && uriNode.getNodeValue() != null)
							referenceExpression.append(SMLCommonUtil.removeLineBreaks(uriNode.getNodeValue()));
					}
				}
			}
			
			String referenceExp = referenceExpression.toString();
			if (referenceExp == null)
			{
				continue;
			}
			
			/* Evaluate the reference */
			try
			{
				URIReference uriReference = new URIReference(referenceExp);
				Node context = null;
				try
				{
					context = uriReference.retrieveDocumentDOM();
				} 
				catch (RemoteRetrievalException e)
				{
					continue;
				}
				if (context == null)
				{
					continue;
				}
				
				String fragmentExpression = uriReference.getFragment();
				if (fragmentExpression == null)
				{
					result.add(context);
					continue;
				}
								
				XPointerExpression xpointerExpression = XPointer.compile(fragmentExpression);
				result.add(xpointerExpression.evaluate(context));				

			} catch (XPointerSyntaxException e)
			{
				exception = e;
			} catch (URISyntaxException e)
			{
				exception = e;
			} catch (BadContextException e)
			{
				exception = e;
			} catch (BadExpressionException e)
			{
				exception = e;
			}
			
			if (exception != null)
				throw new XPathFunctionException(exception); 
		}
		
		return result;
	}

	/**
	 * Evaluates expression to a node object
	 * @param referenceExp uri expression
	 * @return return a node object
	 * @throws XPathFunctionException thrown if an error occurs during evaluation
	 */
	public Object evaluate(String referenceExp) throws XPathFunctionException
	{
		Map domDocuments = (Map) SMLCommonUtil.retrieveDataStructure(DocumentDOMBuilder.ID);		
		SMLIFIdentity identity = (SMLIFIdentity) SMLCommonUtil.retrieveDataStructure(IdentityDataBuilder.ID);
		if (domDocuments == null || identity == null)
			throw new XPathFunctionException(SMLMessages.errorMissingStructure);
		
		Exception exception = null;
		if (referenceExp == null)
		{
			return null;
		}
		
		/* Evaluate the reference */
		try
		{
			URIReference uriReference = new URIReference(referenceExp);
			Node context;
			try
			{
				context = uriReference.retrieveDocumentDOM();
			} 
			catch (RemoteRetrievalException e)
			{
				return null;
			}
			if (context == null)
			{
				return null;
			}
			
			String fragmentExpression = uriReference.getFragment();
			if (fragmentExpression == null)
			{
				return context;
			}
							
			XPointerExpression xpointerExpression = XPointer.compile(fragmentExpression);
			return xpointerExpression.evaluate(context);				

		} catch (XPointerSyntaxException e)
		{
			exception = e;
		} catch (URISyntaxException e)
		{
			exception = e;
		} catch (BadContextException e)
		{
			exception = e;
		} catch (BadExpressionException e)
		{
			exception = e;
		}
		
		if (exception != null)
			throw new XPathFunctionException(exception); 		
		return null;
	}
	/**
	 * Returns an instance of this class
	 * 
	 * @return The instance of this class
	 */
	public static XPathFunction instance()
	{
		if (instance == null)
			instance = new DerefXPathFunction();
		return instance;
	}
	
	
	private static class ArrayNodeList extends ArrayList implements NodeList
	{
		/**
		 * Serial version UID
		 */
		private static final long serialVersionUID = 7992097730072474662L;

		public int getLength()
		{
			return super.size();
		}

		public Node item(int index)
		{
			return (Node)super.get(index);
		}		
	}
}
