/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.smlvalidators;

import java.util.Iterator;
import java.util.Map;

import javax.xml.xpath.XPathFunctionException;

import org.eclipse.cosmos.rm.validation.internal.artifacts.ElementTypeMap;
import org.eclipse.cosmos.rm.validation.internal.artifacts.TargetElementNode;
import org.eclipse.cosmos.rm.validation.internal.artifacts.TargetSchemaNode;
import org.eclipse.cosmos.rm.validation.internal.artifacts.TypeNode;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationOutput;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil;
import org.eclipse.cosmos.rm.validation.internal.core.AbstractSMLValidator;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DocumentDOMBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.ElementTypeMapDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RefTypeElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.SubstitutionBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetInstanceBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetSchemaBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TypeInheritanceDataBuilderImpl;
import org.eclipse.cosmos.rm.validation.internal.reference.DerefXPathFunction;
import org.eclipse.cosmos.rm.validation.internal.util.ParserHelper;
import org.w3c.dom.Element;

/**
 * This class is used to validate the sml:targetElement, sml:targetRequired and sml:targetType sml extensions. 
 * @author sleeloy
 *
 */
public class TargetValidator extends AbstractSMLValidator
{
	
	protected Map targetSchemaBuilderStructure;
	protected Map targetInstanceBuilderStructure;
	protected Map documentDOMBuilderElems;
	protected Map substitionElementElems;
	protected ElementTypeMap elementTypeMapBuilderStructure;
	protected Map inheritance;	
	
	protected IValidationOutput validationLogger; 
	public void initialize(Map validationAttribute)
	{
		super.initialize(validationAttribute);
		DataBuilderRegistry builderRegistry = DataBuilderRegistry.instance();
		
		builderRegistry.registerDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID, new TypeInheritanceDataBuilderImpl());
		builderRegistry.registerDataStructureBuilder(RefTypeElementBuilder.ID, new RefTypeElementBuilder());
		builderRegistry.registerDataStructureBuilder(ElementTypeMapDataBuilder.ID, new ElementTypeMapDataBuilder());
		builderRegistry.registerDataStructureBuilder(TargetSchemaBuilder.ID, new TargetSchemaBuilder());
		builderRegistry.registerDataStructureBuilder(TargetInstanceBuilder.ID, new TargetInstanceBuilder());		
		builderRegistry.registerDataStructureBuilder(SubstitutionBuilder.ID, new SubstitutionBuilder());
		builderRegistry.registerDataStructureBuilder(DocumentDOMBuilder.ID, new DocumentDOMBuilder());				
		
						
	}

	public boolean validate()
	{
		validationLogger = getValidationOutput();
		IDataBuilder targetSchemaBuilder  = DataBuilderRegistry.instance().getDataStructureBuilder(TargetSchemaBuilder.ID);			
		IDataBuilder targetInstanceBuilder  = DataBuilderRegistry.instance().getDataStructureBuilder(TargetInstanceBuilder.ID);			
		IDataBuilder substitionElement  = DataBuilderRegistry.instance().getDataStructureBuilder(SubstitutionBuilder.ID);
		IDataBuilder documentDOMBuilder  = DataBuilderRegistry.instance().getDataStructureBuilder(DocumentDOMBuilder.ID);
		IDataBuilder elementTypeMapBuilder = DataBuilderRegistry.instance().getDataStructureBuilder(ElementTypeMapDataBuilder.ID);
		IDataBuilder typeInheritanceDataBuilder = DataBuilderRegistry.instance().getDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID);

		documentDOMBuilderElems = (Map)documentDOMBuilder.getDataStructure();
		substitionElementElems = (Map)substitionElement.getDataStructure();
		targetSchemaBuilderStructure = (Map)targetSchemaBuilder.getDataStructure();
		targetInstanceBuilderStructure = (Map)targetInstanceBuilder.getDataStructure();
		elementTypeMapBuilderStructure = (ElementTypeMap)elementTypeMapBuilder.getDataStructure();
		inheritance = (Map)typeInheritanceDataBuilder.getDataStructure();
		
		//run target validators
		return validateTargetInstance();
	}
	
	protected boolean validateTargetInstance(){
		Iterator iter = targetInstanceBuilderStructure.keySet().iterator();
		while (iter.hasNext()){
			TargetElementNode targetElementNode = (TargetElementNode)iter.next();
			TargetSchemaNode targetSchemaNode = (TargetSchemaNode)targetSchemaBuilderStructure.get(targetElementNode.getName());
			//check substition group
			if (targetSchemaNode == null){
				Map subGroupMap = (Map)substitionElementElems.get(targetElementNode.getUri());
				if (subGroupMap != null){
					Object subGroup = subGroupMap.get(targetElementNode.getName());
					if (subGroup != null)
						targetSchemaNode = (TargetSchemaNode)targetSchemaBuilderStructure.get(ParserHelper.removeNameSpace((String)subGroup));
				}
				
			}
			
			if ("true".equals(targetSchemaNode.getRequired())){ //$NON-NLS-1$
				//check if target is required
				if (!isURIValid(targetElementNode.getTargetURI())){
					Object[] errorParams = new String[3];
					errorParams[0] = targetElementNode.getName();
					errorParams[1] = targetElementNode.getLine()+""; //$NON-NLS-1$
					errorParams[2] = targetElementNode.getColumn()+""; //$NON-NLS-1$
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.unresolveTargetURI,errorParams));
					return false;									
				}				
			}
			
			//check if we need to check the target element
			if (targetSchemaNode.getTargetElement() != null){
				if (!validateTargetElement(targetElementNode, targetSchemaNode))
					return false;
			}
			if (targetSchemaNode.getTargetType() != null){
				if (!validateTargetType(targetElementNode, targetSchemaNode))
					return false;
			}
			
		}
		return true;		
	}
	protected boolean validateTargetType(TargetElementNode targetElementNode,TargetSchemaNode targetSchemaNode ){

		String targetElementUri = targetElementNode.getTargetURI();
		String targetElementNamespace = targetElementNode.getUri();
		//check if schema node matches instance node
		Object getTargetType = elementTypeMapBuilderStructure.getType(targetElementNode.getUri(), targetElementUri);			
		if ((getTargetType != null)&&(compareElementInstance(targetSchemaNode.getTargetType(), getTargetType))){
			return true;
		}
		else{
			//check inheritance tree
			if (getTargetType != null){
				Map inheritanceMap = (Map)inheritance.get(targetElementNamespace);//getTargetType);
				boolean valid = false;
				if (inheritanceMap != null){
					TypeNode parent = (TypeNode)inheritanceMap.get(getTargetType);
					while (parent != null){
						if (compareElementInstance(parent.getType(),targetSchemaNode.getTargetType())){
							valid = true;
							break;
						}
						parent = (TypeNode)inheritanceMap.get(parent.getType());
						
					}
				}
				if (valid) 
					return true;
			}						
			
			String targetTypeUri = null;
			//the ref could be an alias
			//Search through alias datastructure for actual element
			Object alias = documentDOMBuilderElems.get(targetElementUri);
			if ((alias != null) && (alias instanceof Element)){
				targetTypeUri = ((Element)alias).getNodeName();
				targetElementNamespace  = ((Element)alias).getNamespaceURI();
				
			}
			
			if (targetTypeUri != null)
				getTargetType = elementTypeMapBuilderStructure.getType(targetElementNamespace, targetTypeUri);			
			if ((getTargetType != null) && compareElementInstance(getTargetType,targetSchemaNode.getTargetType())){
				return true;
			}		
			else{
				//check inheritance tree
				if (getTargetType != null){

					Map inheritanceMap = (Map)inheritance.get(targetElementNamespace);
					boolean valid = false;
					if (inheritanceMap != null){
						TypeNode parent = (TypeNode)inheritanceMap.get(getTargetType);
						while (parent != null){
							if (compareElementInstance(parent.getType(),targetSchemaNode.getTargetType())){
								valid = true;
								break;
							}
							parent = (TypeNode)inheritanceMap.get(parent.getType());
							
						}
					}
					if (valid) 
						return true;
				}						
				
			}
			//there could be a substition group
			Object substitionGroup = substitionElementElems.get(targetElementUri);				
			boolean valid = false;
			while (substitionGroup != null){
				getTargetType = elementTypeMapBuilderStructure.getType(targetElementUri,(String)substitionGroup);
				if (compareElementInstance(substitionGroup, targetSchemaNode.getTargetType())){
					valid = true;
					break;
				}			
				substitionGroup = substitionElementElems.get(substitionGroup);
			}
			//if dangling then it's okay
			if (valid || alias == null) return true;					
		}
		Object[] errorParams = new String[5];
		errorParams[0] = targetElementNode.getName();
		errorParams[1] = targetElementNode.getLine() < 0 ? ValidationMessages.commonUnknown : targetElementNode.getLine()+""; //$NON-NLS-1$
		errorParams[2] = targetElementNode.getColumn() < 0 ? ValidationMessages.commonUnknown : targetElementNode.getColumn()+""; //$NON-NLS-1$
		errorParams[3] = targetElementUri;
		errorParams[4] = targetSchemaNode.getTargetType();
		validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetTypeInvalid,errorParams));
		
		return false;			
	}
	
	protected boolean validateTargetElement(TargetElementNode targetElementNode,TargetSchemaNode targetSchemaNode ){

		String targetElementUri = targetElementNode.getTargetURI();
		String targetElementNamespace = targetElementNode.getUri();
		//check if schema node matches instance node
		if (compareElementInstance(targetSchemaNode.getTargetElement(), targetElementUri)){
			return true;
		}
		else{
			//the ref could be an alias
			//Search through alias datastructure for actual element
			Object alias = documentDOMBuilderElems.get(targetElementUri);
			if ((alias != null) && (alias instanceof Element)){
				targetElementUri = ((Element)alias).getNodeName();
				targetElementNamespace  = ((Element)alias).getNamespaceURI();
			}
			if (compareElementInstance(targetElementUri,targetSchemaNode.getTargetElement())){
				return true;
			}				
			//there could be a substition group
			Map substitionGroupMap = (Map)substitionElementElems.get(targetElementNamespace);
			boolean valid = false;
			if (substitionGroupMap != null){
				Object substitionGroup = substitionGroupMap.get(targetElementUri); 
				while (substitionGroup != null){
					if (compareElementInstance(substitionGroup, targetSchemaNode.getTargetElement())){
						valid = true;
						break;
					}			
					substitionGroup = ParserHelper.removeNameSpace((String)substitionGroup);
					substitionGroup = substitionGroupMap.get(substitionGroup);				
				}
			}
			//if dangling then it's okay
			if (valid || alias == null) return true;					
		}
		Object[] errorParams = new String[5];
		errorParams[0] = targetElementNode.getName();
		errorParams[1] = targetElementNode.getLine() < 0 ? ValidationMessages.commonUnknown : targetElementNode.getLine()+""; //$NON-NLS-1$
		errorParams[2] = targetElementNode.getColumn() < 0 ? ValidationMessages.commonUnknown : targetElementNode.getColumn()+""; //$NON-NLS-1$
		errorParams[3] = targetElementUri;
		errorParams[4] = targetSchemaNode.getTargetElement();
		validationLogger.reportError(ValidationMessages.bind(ValidationMessages.invalidTargetInstance,errorParams));
		
		return false;			
	}
	
	protected boolean compareElementInstance(Object elem1, Object elem2){
		return (ParserHelper.removeNameSpace((String)elem1).equals(ParserHelper.removeNameSpace((String)elem2)));
	}
			
	/**
	 * Make sure that the uri can be resolved.
	 * @param uri the uri string to resolve
	 * @return true if the uri can be resolved otherwise false.
	 */
	protected boolean isURIValid(String uri){
		try {
			return (((DerefXPathFunction)DerefXPathFunction.instance()).evaluate(SMLCommonUtil.removeLineBreaks(uri)) != null);
		} catch (XPathFunctionException e) {
			return false;
		}
	}
}
